from pscr_common.aws import get_boto3_resource

from index_registry.utils.config import Config
from index_registry.utils.index_field_helpers import properties_from_key, get_org_from_props, \
    get_size_from_props, get_storage_class_from_props, get_type_from_props


class StatsTable(object):
    """
    This class is responsible for keeping track and creating statistics for all currently hosted and merging indexes per org.
    It is directly plugged in within the lease table. Whenever an index path is being added or deleted in the lease table
    the `add_index` or the `remove_index` methods of this class are being called. The statistics are being kept in a
    dynamoDB table.

    Attributes
    ----------
    config : Config
    table : dynamoDB table as a boto3 resource

    Methods
    -------
    add_index(index_path=)
        Calls the _update_index(index_path=, +1) method
    remove_index(index_path=)
        Calls the _update_index(index_path=, -1) method
    _update_index(index_path=, delta=)
        Gets the necessary information from the index path and makes a call to the _update_metric() method with the extracted
        parameters from the index_path
    _update_metric(org_id=, metric_name=, metric_value=, size_field=, size_value=)
        Makes a call to the DynamoDB table to update the record corresponding to the org_id with the new data

    This is an example for an item in the dynamoDB table
      org_id: primary key (string)
        default_merging_count: int
        default_merging_size: int
        default_optimized_count: int
        default_optimized_size: int
        longTerm_merging_count: int
        longTerm_merging_size: int
        longTerm_optimized_count: int
        longTerm_optimized_size: int

    This is an example of an index path, which should either be deleted or added to the statistics
        "gs://usprod-solr-index-bucket/expire=2021-10-25/class=longTerm/org=7P9F46GN/partition=12/type=optimized/tstamp=179144f32f4/size=53063/index"
    """

    def __init__(self, config: Config):
        self.config = config
        self.table = get_boto3_resource('dynamodb').Table(self.config.get('stats', 'dynamodb_table'))

    def add_index(self, index_path):
        return self._update_index(index_path, +1)

    def remove_index(self, index_path):
        return self._update_index(index_path, -1)

    @staticmethod
    def get_dynamo_fields(index_path):
        index_properties = properties_from_key(index_path)

        org_id = get_org_from_props(props=index_properties)
        index_size = get_size_from_props(props=index_properties)
        index_class = get_storage_class_from_props(props=index_properties)
        index_type = get_type_from_props(props=index_properties)

        if org_id is "" or index_size is "" or ('class=' not in index_path and index_class == 'default') \
                or index_type is "" or index_class is None:
            raise TypeError

        count_metric = index_class + "_" + index_type + "_count"
        size_metric = index_class + "_" + index_type + "_size"

        return org_id, count_metric, size_metric, index_size

    def _update_index(self, index_path, delta):
        org_id, count_metric, size_metric, index_size = self.get_dynamo_fields(index_path)

        self._update_metric(org_id=org_id, count_metric=count_metric, metric_value=delta,
                            size_metric=size_metric, size_value=index_size * delta)

    def _update_metric(self, org_id, count_metric, metric_value, size_metric, size_value):

        return self.table.update_item(
            Key={
                'org_id': org_id
            },
            ExpressionAttributeValues={
                ":metric": metric_value,
                ":size": size_value
            },
            UpdateExpression=f"ADD {count_metric} :metric, {size_metric} :size"
        )

    def scan_table(self):
        result = []
        response = self.table.scan()
        result.extend(response['Items'])

        while 'LastEvaluatedKey' in response:
            response = self.table.scan(
                ExclusiveStartKey=response['LastEvaluatedKey']
            )
            result.extend(response['Items'])

        return result
