import os
import sys
import logging
from urllib.parse import unquote
from dateutil import parser
from enum import Enum

from pscr_common.metrics import PrometheusRegistry
from google.cloud import pubsub_v1

from index_registry.utils.config import Config
from index_registry.engine.index_engine import IndexEngine
from index_registry.utils.index_field_helpers import properties_from_key
from index_registry.engine.searcher_engine import SearcherEngine
from index_registry.engine.org_sqs_engine import OrgSqsEngine
from index_registry.sqs.sqs_queues import get_consumer_name
from index_registry.utils.db import ReadWriteDbPool
from index_registry.utils.metrics_reporter import start_reporter

from pscr_common.log import setup_logging, clear_logging_handlers

_logger = logging.getLogger(__name__)


class GCSEvent(str, Enum):
    OBJECT_CREATED = "OBJECT_FINALIZE"
    OBJECT_DELETED = "OBJECT_DELETE"
    # delete when versioning is enabled
    OBJECT_DELETED_SOFT = "OBJECT_ARCHIVE"


class GCSListener:
    def __init__(
        self,
        config: Config,
        index_engine: IndexEngine,
        org_sqs_engine: OrgSqsEngine,
        searcher_engine: SearcherEngine,
        metrics_registry: PrometheusRegistry,
        trigger_suffix=".trigger",
    ):
        self.config = config
        self.index_engine = index_engine
        self.org_sqs_engine = org_sqs_engine
        self.searcher_engine = searcher_engine
        self.metrics_registry = metrics_registry
        self.trigger_suffix = trigger_suffix
        self.enable_org_sqs = self.config.getboolean(
            "sqs", "enable_org_sqs", fallback=False
        )
        self.env = self.config.get("env", "id")
        self.known_gcs_events = set(GCSEvent)
        self.metrics_prefix = "gcs_events"

    def process_message(self, message):
        """GCS event processing logic"""
        _logger.debug("Got event: %s", message)
        gcs_event = message.attributes

        key = unquote(gcs_event.get("objectId"))
        if not key.endswith(self.trigger_suffix):
            _logger.debug("Skipping event: %s", gcs_event)
            # skip not trigger objects
            message.ack()
            return

        _logger.debug("GCS event: %s", gcs_event)
        event_processed = False
        bucket = unquote(gcs_event.get("bucketId"))
        event_type = gcs_event.get("eventType")

        try:
            if event_type in self.known_gcs_events:
                event_time = parser.parse(gcs_event.get("eventTime"))
                sequencer = ""  # unquote(record['s3']['object']['sequencer'])

                _logger.info(
                    "GCS %s: %s, %s, %s, %s",
                    event_type,
                    bucket,
                    key,
                    sequencer,
                    event_time,
                )

                index_path = self.index_engine.get_gs_uri(
                    bucket, self.index_engine.get_dir_from_trigger(key)
                )

                if event_type == GCSEvent.OBJECT_CREATED:
                    self.searcher_engine.register_placeholder_searcher(index_path)
                    self.index_engine.add_from_s3_uri(
                        index_path, sequencer, last_modified=event_time
                    )
                    self.index_engine.expire_stale_index_partition(index_path)
                    event_processed = True
                elif (
                    event_type == GCSEvent.OBJECT_DELETED
                    or event_type == GCSEvent.OBJECT_DELETED_SOFT
                ):
                    if self.enable_org_sqs:
                        try:
                            properties = properties_from_key(index_path)
                            queue_name = get_consumer_name(properties, self.env)
                            self.org_sqs_engine.org_sqs_delete(
                                properties["org"], queue_name
                            )
                        except KeyError:
                            _logger.error(
                                f"Bad parsing of {index_path} skipping sqs deletion"
                            )
                    else:
                        _logger.info("Org SQS is disabled")
                    self.index_engine.delete_index(
                        index_path, sequencer, last_modified=event_time
                    )
                    event_processed = True
            else:
                _logger.warning("Unknown event: %s", event_type)
                event_type = "UNKNOWN"

        except Exception as ex:
            _logger.error(f"Unexpected exception: {str(ex)}", exc_info=True)
        else:
            # acknowledge if no exceptions
            message.ack()
        finally:
            event_processed_str = "SUCCESS" if event_processed else "FAIL"
            self.metrics_registry.meter(
                f"{self.metrics_prefix}.{event_type}.{event_processed_str}"
            ).mark()

    def listen(self):
        """Start PubSub subscription listener"""

        def _callback(message):
            """Call this on each message"""
            self.process_message(message)

        subscription_name = "projects/{project_id}/subscriptions/{sub}".format(
            project_id=self.config.get("gcp", "service_project_id"),
            sub=self.config.get("pubsub", "indexes_subscription_name"),
        )

        with pubsub_v1.SubscriberClient() as subscriber:
            future = subscriber.subscribe(subscription_name, _callback)
            future.result()


def main():
    config = Config.create()
    clear_logging_handlers()
    setup_logging(
        level=config.get("log", "level"),
        svc_version=os.environ.get("CONTAINERVERSION", "UNKNOWN"),
        env_id=os.environ.get("ENV_ID", "UNKNOWN"),
        svc_id="TODO",
    )
    # The GCP package has a lot of noisy debug logs, so have a separate config for that
    logging.getLogger("google").setLevel(config.get("log.gcp", "level"))

    metrics_registry = PrometheusRegistry()
    start_reporter(metrics_registry)

    db_pool = ReadWriteDbPool(config)

    trigger_suffix = os.environ.get("TRIGGER_SUFFIX", ".trigger")
    index_engine = IndexEngine(trigger_suffix, config, db_pool=db_pool)
    org_sqs_engine = OrgSqsEngine(config)
    searcher_engine = SearcherEngine(config, db_pool=db_pool)

    gcs_listner = GCSListener(
        config,
        index_engine,
        org_sqs_engine,
        searcher_engine,
        metrics_registry,
        trigger_suffix,
    )
    gcs_listner.listen()


if __name__ == "__main__":
    sys.exit(main())
