from collections import defaultdict
import datetime
import logging
import os, time
from sqlalchemy import func, null
from sqlalchemy.exc import ProgrammingError

from index_registry.engine.searcher_engine import SearcherEngine
from index_registry.utils.db import ReadWriteDbPool
from index_registry.sql.models import Index, Lease, Searcher
from pscr_common.utils import get_utcnow_with_tzinfo
from pscr_common.metrics import get_reporter, PrometheusRegistry
from pscr_common.metadata import MetadataProvider
from index_registry.mongodb.leases_db_interface import DbManager as LeasesDbManager

_logger = logging.getLogger(__name__)


class Metrics(object):
    def __init__(self, config, db_pool: ReadWriteDbPool):
        self.db_pool = db_pool
        # this is cron, so send metrics directly
        self._registry = PrometheusRegistry(batch_job=True)
        self.lease_table = LeasesDbManager(config)
        self.config = config
        self.reporter = self.get_reporter(self.config, self._registry)

    def collect(self):
        # Collect index counts
        for index_prefix, count in self._get_index_counts().items():
            _logger.debug("Index Prefix: %s, Count: %s", index_prefix, count)
            self._registry.gauge("indexes", extra_labels={"index_prefix": index_prefix}).set_value(count)

        # Collect lease counts
        for index_prefix, version_counts in self._get_lease_counts().items():
            for version, count in version_counts.items():
                _logger.debug(
                    "Lease Index Prefix: %s, Container Version: %s, Count: %s",
                    index_prefix,
                    version,
                    count,
                )
                self._registry.gauge(
                    "leases", extra_labels={"index_prefix": index_prefix, "index_version": version}
                ).set_value(count)

        # Collect active searchers
        for version, count in self._get_active_searchers().items():
            _logger.debug("Active Searcher Version: %s, Count: %s", version, count)
            self._registry.gauge("searchers.active", extra_labels={"index_version": version}).set_value(count)

        # Collect expired searchers
        for version, count in self._get_expired_searchers().items():
            _logger.debug("Expired Searcher Version: %s, Count: %s", version, count)
            self._registry.gauge("searchers.expired", extra_labels={"index_version": version}).set_value(count)

        self._registry.gauge("leases.total.merge").set_value(
            self.lease_table.get_merging_count()
        )
        self._registry.gauge("leases.total.optimized").set_value(
            self.lease_table.get_optimized_count()
        )
        self._registry.gauge("leases.leased.merge").set_value(
            self.lease_table.get_merging_leased()
        )
        self._registry.gauge("leases.leased.optimized").set_value(
            self.lease_table.get_optimized_leased()
        )
        self._registry.gauge("leases.hosted.merge").set_value(
            self.lease_table.get_merging_hosted()
        )
        self._registry.gauge("leases.hosted.optimized").set_value(
            self.lease_table.get_optimized_hosted()
        )

    def publish(self):
        _logger.debug("Publishing metrics...")
        if self.reporter:
            self.reporter.report_now(self._registry, "maintenance")
            # TODO:  Added this sleep because wavefront metrics never appear without, should re-evaluate after moving to aiops of we still need this.
            time.sleep(120)
            _logger.debug("Done Publishing metrics")
        else:
            _logger.warning("No reporter, metrics did not publish")

    @staticmethod
    def get_reporter(config, registry):
        env = os.environ.get("ENV_ID", "unknown")
        env_meta = MetadataProvider().get_metadata()
        version = config.get("service", "version")
        if env_meta.runtime != "lambda":
            version = env_meta.image.rsplit(":", maxsplit=1)[-1]

        _logger.debug(
            f"Metrics Reporter, env:{env}, task_id:{env_meta.task_id}, version:{version}, pid:{str(os.getpid())}"
        )
        reporter = get_reporter(
            registry,
            env,
            env_meta.task_id,
            "index_registry",
            version,
            str(os.getpid()),
            service_name_as_prefix=True,
        )

        return reporter

    def _get_index_counts(self):
        prefix_counts = {"raw": 0, "merging": 0, "optimized": 0}

        with self.db_pool.scoped_session(read_only=True) as session:
            try:
                results = (
                    session.query(
                        Index.prefix,
                        func.count(Index.index_path).label("index_path_count"),
                    )
                    .filter(Index.date_removed == null())
                    .group_by(Index.prefix)
                    .all()
                )
                for record in results:
                    prefix_counts[
                        self._get_safe_index_label(record.prefix)
                    ] = record.index_path_count
            except ProgrammingError:
                _logger.debug("Database not ready while getting index counts")

        return prefix_counts

    def _get_lease_counts(self, now: datetime.datetime = None):
        now = now or get_utcnow_with_tzinfo()

        lease_counts = defaultdict(dict)
        with self.db_pool.scoped_session(read_only=True) as session:
            try:
                results = (
                    session.query(
                        Index.prefix,
                        Lease.container_version,
                        func.count(Lease.index_path).label("index_path_count"),
                    )
                    .join(Lease)
                    .filter(Lease.expiration > now)
                    .filter(Index.date_removed == null())
                    .group_by(Index.prefix, Lease.container_version)
                    .all()
                )
                for record in results:
                    lease_counts[self._get_safe_index_label(record.prefix)][
                        self._get_safe_index_label(record.container_version)
                    ] = record.index_path_count
            except ProgrammingError:
                _logger.debug("Database not ready while getting lease counts")

        for prefix in ("merging", "optimized"):
            if prefix not in lease_counts:
                lease_counts[prefix]["NA"] = 0

        return lease_counts

    def _get_active_searchers(self, now: datetime.datetime = None):
        now = now or get_utcnow_with_tzinfo()

        searchers = {}
        with self.db_pool.scoped_session(read_only=True) as session:
            try:
                results = (
                    session.query(
                        Searcher.container_version,
                        func.count(Searcher.searcher_id).label("searcher_count"),
                    )
                    .filter(Searcher.expiration > now)
                    .filter(
                        Searcher.container_version
                        != SearcherEngine.PLACEHOLDER_CONTAINER_VERSION
                    )
                    .group_by(Searcher.container_version)
                    .all()
                )
                for record in results:
                    searchers[
                        self._get_safe_index_label(record.container_version)
                    ] = record.searcher_count
            except ProgrammingError:
                _logger.debug("Database not ready while getting active searchers")

        return searchers

    def _get_expired_searchers(self, now: datetime.datetime = None):
        now = now or get_utcnow_with_tzinfo()

        searchers = {}
        with self.db_pool.scoped_session(read_only=True) as session:
            try:
                results = (
                    session.query(
                        Searcher.container_version,
                        func.count(Searcher.searcher_id).label("searcher_count"),
                    )
                    .filter(Searcher.expiration < now)
                    .filter(
                        Searcher.container_version
                        != SearcherEngine.PLACEHOLDER_CONTAINER_VERSION
                    )
                    .group_by(Searcher.container_version)
                    .all()
                )
                for record in results:
                    searchers[
                        self._get_safe_index_label(record.container_version)
                    ] = record.searcher_count
            except ProgrammingError:
                _logger.debug("Database not ready while getting expired searchers")

        return searchers

    @staticmethod
    def _get_safe_index_label(label: str):
        return label.replace(".", "-")
