import json
import boto3

from typing import Dict, Any
from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy import text
from sqlalchemy.engine.url import URL

from index_registry.utils.config import Config


# Try getting the password in the following order - from secrets managers, from ssm, from raw
def get_password(config: Config, config_section: str):
    password_secrets_store_key = config.get(config_section, 'password_secrets_manager_key')
    password_ssm_key = config.get(config_section, 'password_ssm_key')
    raw_password = config.get(config_section, "password")

    return get_password_from_secret_store(password_secrets_store_key) if password_secrets_store_key \
        else get_password_from_ssm(password_ssm_key) if password_ssm_key \
        else raw_password


def get_password_from_secret_store(key):
    secrets_store = boto3.client("secretsmanager")
    response = secrets_store.get_secret_value(SecretId=key)['SecretString']
    return json.loads(response)["password"]


def get_password_from_ssm(key):
    ssm = boto3.client("ssm")
    response = ssm.get_parameter(Name=key, WithDecryption=True)

    return response['Parameter']['Value']


def get_orm_connect_config(config: Config, read_only=False) -> Dict[str, Any]:
    if read_only and config.getboolean('db.readonly', 'enabled', fallback=False):
        section = "db.readonly"
    else:
        section = 'db.readwrite'

    return {
        'host': config.get(section, 'host'),
        'port': config.getint(section, 'port'),
        'database': config.get(section, 'name'),
        'username': config.get(section, 'user'),
        'password': get_password(config, section),
        'connect_timeout': config.getint(section, 'connect_timeout')
    }


def truncate_all(db_pool):
    with db_pool.scoped_session() as session:
        session.execute(text("TRUNCATE TABLE indexes CASCADE"))
        session.execute(text("TRUNCATE TABLE leases"))
        session.execute(text("TRUNCATE TABLE searchers"))


def get_engine(config: Config, read_only=False):
    db_config = get_orm_connect_config(config, read_only=read_only)
    connect_timeout = db_config.pop('connect_timeout', None)

    return create_engine(URL.create('postgresql', **db_config), connect_args={'connect_timeout': connect_timeout}, pool_size=5, pool_recycle=3500)


class ReadWriteDbPool(object):
    def __init__(self, config, seed=None):
        self.seed = seed
        self.rw_engine = get_engine(config, False)
        self.rw_session_maker = sessionmaker(bind=self.rw_engine)

        if config.getboolean('db.readonly', 'enabled', fallback=False):
            self.ro_engine = get_engine(config, True)
            self.ro_session_maker = sessionmaker(bind=self.ro_engine)
        else:
            self.ro_engine = self.rw_engine
            self.ro_session_maker = self.rw_session_maker

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self.close()

    @contextmanager
    def scoped_session(self, read_only=False):
        session = self.ro_session_maker() if read_only else self.rw_session_maker()
        if self.seed is not None:
            session.execute(f'SELECT setseed({self.seed})')

        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def close(self):
        self.rw_engine.dispose()
        self.ro_engine.dispose()
